function model = gpfnarx(data)
%GPFNARX creates a Gaussian process model of a nonlinear dynamical system.
%
% The input argument DATA needs to be a structure containing the input
% signals in DATA.u and the outputs in DATA.y. An iddata object is
% acceptable.
%
% The output MODEL contains the Gaussian process model of the dynamical
% system whose mean can be easily evaluated by using the function:
%   MODEL.predictFITC = @(ytest,utest)
%
% See preprint: Roger Frigola and Carl Rasmussen, Automated Bayesian System
% Identification with NARX Models, 2013.
%
% Please contact me at rf342@cam.ac.uk with any queries.
%
% Copyright (c) by Roger Frigola, March 2013. 


% Do some checks
if size(data.y,2) > 1 || size(data.u,2) > 1
    error('Multidimensional signals not yet implemented.')
elseif exist('gp.m','file') ~= 2
    error('The GPML toolbox is required: http://www.gaussianprocess.org/gpml/code')
end

% Setup model order and FITC parameters
model.numPrevInputs = 10;
model.numPrevOut = 10;
model.numTrainPoints = 2^9;
model.numValidationPoints = 2^10;

% Split data into a subset for training/prediction and the rest for validation
aux = randperm(length(data.y) - max(model.numPrevInputs, model.numPrevOut) - 1);
iiTrainSubset = aux(1:model.numTrainPoints);
iiValidationSubset = aux(model.numTrainPoints+1 : ...
    min(model.numTrainPoints+model.numValidationPoints, length(aux)));

% Initialise hyperparameters
hyp0 = InitialiseHyperparameters(data, model);
wn0 = 10; % This is a transformed w 
[XFilt, yFilt, ii] = AssembleRegressors(MyFilt(data.y,wn0), data.u, model);
yNotFilt = data.y(ii);
hyp0 = minimize(hyp0, @FastMarginalLikelihood, -30, ...
    XFilt(iiTrainSubset,:), yNotFilt(iiTrainSubset,:));
theta_wn0 = [hyp0.cov; hyp0.lik; wn0];

% Just optimise the filter to get a better initialisation
costwn = @(wn) WrapMarginalLikelihood(data.y, data.u, iiTrainSubset, ...
    Theta2Hyp(theta_wn0),wn,model);
wn0 = fminbnd(costwn,-10,10); % This is a transformed w
theta_wn0 = [hyp0.cov; hyp0.lik; wn0];

% Jointly optimise hypers and filter
cost = @(theta_wn) WrapMarginalLikelihood(data.y, data.u, iiTrainSubset, ...
    Theta2Hyp(theta_wn),theta_wn(end),model);
theta_wnOpt = minimize(theta_wn0, cost, -70);

% Package all results into model
model.hyp.cov = theta_wnOpt(1:end-2);
model.hyp.lik = theta_wnOpt(end-1);
model.wn = theta_wnOpt(end);

% Precompute stuff and ship it with the model
[~, model.alpha, model.U, model.XTrain] = ValidationError(data.y, data.u, iiTrainSubset, ...
    iiValidationSubset,Theta2Hyp(theta_wnOpt),theta_wnOpt(end),model);
% model.predict = @(ytest,utest) covSEard(model.hyp.cov, ...
%     AssembleRegressors(MyFilt(ytest,model.wn), utest, model), model.XTrain) * model.alpha;

[XFilt, yFilt, ii] = AssembleRegressors(MyFilt(data.y,model.wn), data.u, model);
yNotFilt = data.y(ii);
XTrain = XFilt(iiTrainSubset,:);
cov = {@covFITC, {@covSEard}, XTrain};
model.hyp.mean = [];

alpha = FITCalpha(model.hyp, {@meanZero}, cov, @likGauss, XFilt, yNotFilt);
model.predictFITC = @(ytest,utest) covSEard(model.hyp.cov, ...
    AssembleRegressors(MyFilt(ytest,model.wn), utest, model), ...
    model.XTrain) * alpha;        


end




%%%%%%%%%%%%%%%%%%%%         PRIVATE FUNCTIONS         %%%%%%%%%%%%%%%%%%%%


function [p, dpdthetawn] = WrapMarginalLikelihood(y,u,iiTrain,hyp,wn,model)

[XFilt, yFilt, ii] = AssembleRegressors(MyFilt(y,wn), u, model);
yNotFilt = y(ii);
XTrain = XFilt(iiTrain,:);
yTrain = yNotFilt(iiTrain,:);  % Train with not filtered!

[p, dpdtheta] = FastMarginalLikelihood(hyp, XTrain, yTrain);

% Brute force derivative wrt wn
pert = 1e-7;
[XFilt2, yFilt2, ii] = AssembleRegressors(MyFilt(y,wn+pert), u, model);
XTrain2 = XFilt2(iiTrain,:);
ppert = FastMarginalLikelihood(hyp, XTrain2, yTrain);
dpdwn = (ppert - p) / pert;

% % Compute derivative wrt wn. It would be better to use the chol
% % already computed in infExact...
% [n, D] = size(XTrain);
% K = feval(@covSEard, hyp.cov, XTrain);                      % evaluate covariance matrix
% sn2 = exp(2*hyp.lik);                               % noise variance of likGauss
% L = chol(K/sn2+eye(n));               % Cholesky factor of covariance with noise
% alpha = solve_chol(L,yTrain)/sn2;
% Q = solve_chol(L,eye(n))/sn2 - alpha*alpha';    % precompute for convenience
% for i = 1:numel(hyp.cov)
%     dnlZ.cov(i) = sum(sum(Q.*GetdKdwn()))/2;
% end

dpdthetawn = [dpdtheta.cov; dpdtheta.lik; dpdwn];

end



function yFilt = MyFilt(y,wn)

wnActual = 1 / (1 + exp(-wn)); 
 % wnActual needs to be between > 0 and < 1
if wnActual >= 1
    yFilt = y;
    return
elseif wnActual <= 0
    yFilt = 0 * y;
    return
end

[b,a] = butter(2, wnActual);
yFilt = filtfilt(b,a,y);

end



function [err, varargout] = ValidationError(y,u,iiTrain,iiVal,hyp,wn,model)

% Assemble test and train matrices
[XFilt, yFilt, ii] = AssembleRegressors(MyFilt(y,wn), u, model);
yNotFilt = y(ii);
XTrain = XFilt(iiTrain,:);
yTrain = yFilt(iiTrain,:);
XVal = XFilt(iiVal,:);

bFITC = 0;
if ~bFITC
    Kstar = covSEard(hyp.cov, XVal, XTrain);
    Kn = covSEard(hyp.cov, XTrain);
    for i = 1:size(Kn,1)
        Kn(i,i) = Kn(i,i) + exp(hyp.lik)^2;
    end
    U = chol(Kn);
    alpha = solve_chol(U, yTrain);
    %muTest = Kstar * (Kn \ yTrain);
    muTest = Kstar * alpha;
else
    numInducingPoints = 100;
    covfun = {@covFITC, {@covSEard}, XTrain(1:numInducingPoints,:)};
    muTest = gp(hyp, @infFITC, [], covfun, @likGauss, XFilt, yFilt, XVal);
end

%err = sqrt(mean((yNotFilt(iiVal) - muTest).^2));
err = yNotFilt(iiVal) - muTest;

if nargout > 1
    varargout{1} = alpha;
    varargout{2} = U;
    varargout{3} = XTrain;
end

end



function [X,Y,ii] = AssembleRegressors(y, u, model)
% Assembles regressors and potentially targets (Y).
% Works for one dimensional signals

ii = max(model.numPrevInputs, model.numPrevOut)+1:numel(y);

Xy = y(bsxfun(@minus, repmat(ii',1,model.numPrevOut), 1:model.numPrevOut));
Xu = u(bsxfun(@minus, repmat(ii',1,model.numPrevInputs+1), 0:model.numPrevInputs));

X = [Xy Xu];

if nargout > 1
    Y = y(ii,:);
end
end



function hyp = Theta2Hyp(theta_wn)

theta = theta_wn(1:end-1);
hyp.cov = theta(1:end-1);
hyp.lik = theta(end);
end



function hyp0 = InitialiseHyperparameters(data, model)

[X,y] = AssembleRegressors(data.y, data.u, model);

% Learn hyper-parameters
hyp0.cov = log([3*std(X)'; std(y)]);
% hyp0.cov(~isfinite(hyp0.cov)) = log(2);
hyp0.lik = log(std(y) / 20);

end



function [alpha, L] = FITCalpha(hyp, mean, cov, lik, x, y)
% This is derived from infFITC in the GPML toolbox.


if iscell(lik), likstr = lik{1}; else likstr = lik; end
if ~ischar(likstr), likstr = func2str(likstr); end
if ~strcmp(likstr,'likGauss')               % NOTE: no explicit call to likGauss
  error('Inference with inFITC only possible with Gaussian likelihood.');
end
cov1 = cov{1}; if isa(cov1, 'function_handle'), cov1 = func2str(cov1); end
if ~strcmp(cov1,'covFITC'); error('Only covFITC supported.'), end    % check cov

[diagK,Kuu,Ku] = feval(cov{:}, hyp.cov, x);         % evaluate covariance matrix
m = feval(mean{:}, hyp.mean, x);                          % evaluate mean vector
[n, D] = size(x); nu = size(Kuu,1);

sn2  = exp(2*hyp.lik);                              % noise variance of likGauss
snu2 = max(1e-6*sn2, 1e-8);                              % hard coded inducing inputs noise
Luu  = chol(Kuu+snu2*eye(nu));                         % Kuu + snu2*I = Luu'*Luu
V  = Luu'\Ku;                                     % V = inv(Luu')*Ku => V'*V = Q
g_sn2 = diagK + sn2 - sum(V.*V,1)';          % g + sn2 = diag(K) + sn2 - diag(Q)
Lu = chol(eye(nu) + (V./repmat(g_sn2',nu,1))*V');  % Lu'*Lu=I+V*diag(1/g_sn2)*V'
r  = (y-m)./sqrt(g_sn2);
be = Lu'\(V*(r./sqrt(g_sn2)));
iKuu = solve_chol(Luu,eye(nu));                       % inv(Kuu + snu2*I) = iKuu
alpha = Luu\(Lu\be);                      % return the posterior parameters
if nargout > 1
    L  = solve_chol(Lu*Luu,eye(nu)) - iKuu;                    % Sigma-inv(Kuu)
end
%post.sW = ones(n,1)/sqrt(sn2);           % unused for FITC prediction  with gp.m

end



function [nlZ dnlZ] = FastMarginalLikelihood(hyp,x,y)
% This has been derived from infExact and covSEard in the GPML toolbox. The
% goal is to make computation faster for the particular case of zero mean
% and covSEard covariance function.

cov = @covSEard;

[n, D] = size(x);
K = feval(cov, hyp.cov, x);                      % evaluate covariance matrix
sn2 = exp(2*hyp.lik);                               % noise variance of likGauss
L = chol(K/sn2+eye(n));               % Cholesky factor of covariance with noise
alpha = solve_chol(L,y)/sn2;

if nargout > 0                               % do we want the marginal likelihood?
    nlZ = (y)'*alpha/2 + sum(log(diag(L))) + n*log(2*pi*sn2)/2;  % -log marg lik
    if nargout > 1                                        % do we want derivatives?
        dnlZ = hyp;                                 % allocate space for derivatives
        Q = solve_chol(L,eye(n))/sn2 - alpha*alpha';    % precompute for convenience
        
        ell = exp(hyp.cov(1:D));                               % characteristic length scale
        sf2 = exp(2*hyp.cov(D+1));                                         % signal variance
        
        % precompute squared distances
        K = sq_dist(diag(1./ell)*x');
        K = sf2*exp(-K/2);
        for i = 1:numel(hyp.cov)
            if i<=D                                              % length scale parameters
                dK = K.*sq_dist(x(:,i)'/ell(i));
            elseif i==D+1                                            % magnitude parameter
                dK = 2*K;
            else
                error('Unknown hyperparameter')
            end
            dnlZ.cov(i) = sum(sum(Q.*dK))/2;
        end
        dnlZ.lik = sn2*trace(Q);
    end
end
end